/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * File: p4im.cpp
 *
 * Implements p4im, a small command-line client and plugin for P4. p4im
 * supports IRC-style instant messaging. It can display the users that are
 * logged on and send messages to a specific user or all users.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */


#include "p4client.h"
#include "p4plugin.h"
#include "p4error.h"
#include "../libp4/settings.h"

#ifdef _WIN32
#include <sys/timeb.h>
#include <Winsock2.h>
#else
#include <sys/time.h>
#include <unistd.h>
#endif

#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>
#include <algorithm>
#include <ctime>
#include <cstdlib>
#include <cassert>
#include <errno.h>
#include <sys/types.h>
using namespace std;


/*****************************************************************************
 * FORWARD DECLARATIONS
 *****************************************************************************/
struct cmd_t;

/*****************************************************************************
 * CONSTANTS
 *****************************************************************************/
const string kp4im_welcome =
  "Welcome to p4im. Type '/help' for a command list.";
const string kp4im_goodbye = "Goodbye!";
const string kp4im_prompt = ":p4im> ";
const ushort kp4im_pluginid = 1001;
const uint kstdin = 0;    // the file descriptor for stdin
const int kmsg_max_len = 256;


/*****************************************************************************
 * GLOBALS
 *****************************************************************************/
string guser = csettings::instance().get<string>("local_username");
vector<cmd_t> gcmds;
vector<ushort> plugins;


/*****************************************************************************
 * PROTOTYPES
 *****************************************************************************/
void *plugin_main(void *arg);

// the user commands
void quit_cmd     (const string &unused1 = "",   const string &unused2 = "");
void names_cmd    (const string &plugin_id = "", const string &unused2 = "");
void conns_cmd    (const string &unused1 = "",   const string &unused2 = "");
void me_cmd       (const string &msg = "",       const string &unused = "");
void msg_cmd      (const string &user = "",      const string &msg = "");
void help_cmd     (const string &unused1 = "",   const string &unused2 = "");
void broadcast_cmd(const string &msg = "",       const string &unused = "");
void set_cmd      (const string &key = "",       const string &value = "");
void open_cmd     (const string &user = "",      const string &addr_port = "");
void load_cmd     (const string &plugin_id = "", const string &unused2 = "");
void unload_cmd   (const string &plugin_id = "", const string &unused2 = "");
void plugins_cmd  (const string &unused1 = "",   const string &unused2 = "");

// helper functions
void error(const string &err);
void parse(const string &line);


/*****************************************************************************
 * COMMANDS
 *****************************************************************************/
typedef void (*cmd_fn_t)(const string &arg1 = string(""),
                         const string &arg2 = string(""));

struct cmd_t {
  cmd_t(const string &cmd, const string &desc, cmd_fn_t fn)
    : cmd(cmd), desc(desc), fn(fn) {}
  cmd_t(const char *cmd, const char *desc, cmd_fn_t fn)
    : cmd(cmd), desc(desc), fn(fn) {}

  string cmd;
  string desc;
  cmd_fn_t fn;
};

void populate_cmds()
{
  // the commands - to run these commands, precede them with '/' and, if
  // 'me' or 'msg' or 'set', follow them with arguments

  cmd_t quit("quit", "Quits p4im.", quit_cmd);
  gcmds.push_back(quit);

  cmd_t me("me", "Like /me in IRC. *NOT IMPLEMENTED*.", me_cmd);
  gcmds.push_back(me);

  cmd_t msg("msg", "Sends the message only to the given user.", msg_cmd);
  gcmds.push_back(msg);

  cmd_t names("names", "Prints the names of all users currently online. "
              "If a plugin id is passed as the argument, the users running "
              "that plugin are printed.", names_cmd);
  gcmds.push_back(names);

  cmd_t conns("conns", "Prints the info about the users' computers to "
              "which we are currently connected.  It prints the remote "
              "user's name followed by IP address, port, and status "
              "of the connection.", conns_cmd);
  gcmds.push_back(conns);

  cmd_t help("help", "Prints a list of all supported commands.", help_cmd);
  gcmds.push_back(help);

  cmd_t set("set", "Prints a list of all settings and their values.", set_cmd);
  gcmds.push_back(set);

  cmd_t open("open", "Opens a new connection to a P4 client. Takes a username"
             "and a server and port in <addr>:<port> notation. For example:\n"
             "\tyourmom:p4im> open ryan blaize.stanford.edu:4002" , open_cmd);
  gcmds.push_back(open);

  cmd_t load("load", "Loads a plugin id. P4IM will add this plugin id to the "
             "current list of plugins that this client is running. Note that "
             "P4IM will only receive IM messages at plugin ID 1001, though.",
             load_cmd);
  gcmds.push_back(load);

  cmd_t unload("unload", "Unloads a plugin id. P4IM will remove this plugin "
               "id from the current list of plugins this client is running.",
               unload_cmd);
  gcmds.push_back(unload);

  cmd_t plugins("plugins", "Lists the plugin ids that are currently loaded.",
                plugins_cmd);
  gcmds.push_back(plugins);
}


/*****************************************************************************
 * PLUGIN_MAIN
 *****************************************************************************/
int main(int argc, char **argv)
{
  clock_t start = clock();

  // start up the network
  uint boot_ret = p4_bootstrap();
  if (boot_ret == kerr_boot_init_failed) {
    cerr << "Could not bootstrap" << endl;
    return -1;
  } else if (boot_ret == kerr_couldnt_connect) {
    cerr << "Not connected.  Listening for connections..." << endl;
  }

  clock_t booted = clock();
  p4_plugin_loaded(kp4im_pluginid);

  // measure total load time
  clock_t done = clock();
  float boottime = static_cast<float>(booted - start) / CLOCKS_PER_SEC;
  float loadtime = static_cast<float>(done - booted) / CLOCKS_PER_SEC;
  cout << "Backend boot time " << boottime + loadtime << " seconds ("
       << boottime << " + " << loadtime << ") (grossly inaccurate!)" << endl;


  // start p4im!
  void *ret = plugin_main(0);

  p4_quit();
  return (ret) ? -1 : 0;
}


/** plugin_main
 *
 * The main entry point for the plugin.
 */
void *plugin_main(void *arg)
{
  char  msg[kmsg_max_len];
  fd_set  stdinfd;
  struct timeval timeout;

  // init
  populate_cmds();
  plugins.push_back(kp4im_pluginid);

  cout << kp4im_welcome << endl << guser << kp4im_prompt << flush;

  // main program loop
  while (true) {

    // reset the timeout each iteration cause select changes it
    timeout.tv_sec = 0;
    timeout.tv_usec = 100000;

    // is there user input?
    FD_ZERO(&stdinfd);
    FD_SET(kstdin, &stdinfd);
    int ret  = select(kstdin + 1, &stdinfd, 0, 0, &timeout);
    if (ret < 0 && errno != 0) {
      cerr << "Select returned error number " << errno << endl;
      error("select error");
    }

    if (FD_ISSET(kstdin, &stdinfd)) {  // yes, there is!
      cin.setf(ios::skipws);
      if (!cin.getline(msg, kmsg_max_len))
        quit_cmd();;                   // received EOF, say goodbye
      msg[min(cin.gcount(), kmsg_max_len - 1)] = '\0';

      if (string(msg) != "") {
        parse(msg);
        cout << endl;
      }
      cout << guser << kp4im_prompt << flush;
    }

    // is there network data?
    int data_len;

    p4_peek(kp4im_pluginid, &data_len);
    if (data_len > 0) {              // yes, there is!
      int written;
      bool encrypted;
      char user[kp4_username_len];

      uint ret = p4_receive(kp4im_pluginid, user, msg, kmsg_max_len,
                  &written, &encrypted);
      if (ret != 0)
        cerr << "p4_receive error " << ret << endl;
      assert(!encrypted);
      assert(msg[written - 1] == '\0');
      cout << endl << msg << endl << guser << kp4im_prompt << flush;
    }
  }
}



/*****************************************************************************
 * COMMAND FUNCTIONS
 *****************************************************************************/

/** quit_cmd
 *
 * Exits the program. Called when a user enters "/quit".
 */
void quit_cmd(const string &unused1, const string &unused2)
{
  cout << kp4im_goodbye << endl;
  p4_quit();
  exit(0);
}


/** names_cmd
 *
 * Prints out the usernames of all users logged onto the network, one username
 * per line. Called when a user enters "/names". Equivalent to /names in IRC.
 *
 * If the first argument isn't blank, it is interpreted as a plugin id and the
 * names cmd prints all of the names for that plugin. If the plugin id is 0,
 * the names of *all* of the users connected to the network are printed.
 */
void names_cmd(const string &plugin_id, const string &unused2)
{
  int numusers;
  char **names;

  // get plugin id
  ushort id = kp4im_pluginid;
  if (plugin_id != "") {
    stringstream stream;
    stream << plugin_id;
    stream >> id;
  }

  // get number of names
  uint ret = p4_get_num_users(id, &numusers);
  assert(!ret);
  if (numusers == 0)
    return;

  // get names
  names = new char*[numusers];
  assert(names);
  for (int i = 0; i < numusers; i++) {
    names[i] = new char[kp4_username_len];
    assert(names[i]);
  }

  ret = p4_get_users(id, reinterpret_cast<char **>(names), numusers,
                     &numusers);
  assert(!ret);

  // print names
  for (int i = 0; i < numusers; i++)
    cout << names[i] << endl;


  // clean up
  // TODO: leaks memory cause numusers could be different after
  // p4_get_users call
  for (int i = 0; i < numusers; i++)
    delete[] names[i];
  delete[] names;
}

/** conns_cmd
 *
 * Prints the current connections for the backend.
 */
void conns_cmd(const string &unused1, const string &unused2)
{
  int num_conns, num_written;
  p4_conn_info *conns;

  p4_get_num_conns(&num_conns);
  
  // allocate memory for the conns
  conns = new struct p4_conn_info[num_conns];

  p4_get_conns(conns, num_conns, &num_written);

  ofstream dot_file("p4network.dot", ios::trunc);

  dot_file << "graph net { " << endl;  
  dot_file << "size = \"4, 4\"; center = true;" << endl;
  dot_file << guser << "[shape=box]" << endl;

  // print them
  for (int i = 0; i < num_written; i++) {
    dot_file << guser << " -- " << conns[i].username << ";" << endl;

    cout << conns[i].username << " " << conns[i].addr << " "
         << conns[i].port << " status: ";
    switch (conns[i].status) {
    case eopening: cout << "opening"; break;
    case epreamble_wait: cout << "preamble_wait"; break;
    case epreamble_ack_wait: cout << "preamble_ack_wait"; break;
    case econnected: cout << "connected"; break;
    case efailed: cout << "failed"; break;
    case eclosed: cout << "closed"; break;
    default: cout << "unknown! " << conns[i].status;
    }
    cout << endl;
  }
  delete[] conns;
  dot_file << "}" << endl;
}
  

  
void me_cmd(const string &msg, const string &unused)
{
  cout << "me_cmd not implemented" << endl;
}


/** msg_cmd
 *
 * Sends a message directly to the given user. Called when a user enters
 * "/msg". Equivalent to /msg in IRC.
 */
const uint    kerr_user_offline = 2;

void msg_cmd(const string &user, const string &msg)
{
  // build message
  string tagged = guser + "->" + user + ": " + msg;

  // send message
  uint ret = p4_send(kp4im_pluginid, user.c_str(), tagged.c_str(),
                     tagged.length() + 1, false);

  // handle errors
  if (ret == 0)
    cout << tagged;
  else if (ret == kerr_user_offline)
    cout << user << " is not logged on.";
  else
    cout << "p4_send error " << ret << endl;
}


/** help_cmd
 *
 * Prints out a list of the supported commands. Called when a user enters
 * "/help".
 */
void help_cmd(const string &unused1, const string &unused2)
{
  cout << "To send a message to all online users, enter it at the prompt."
       << endl << "These are all the commands currently supported by p4im."
       << endl;

  for (uint i = 0; i < gcmds.size(); i++)
    cout << "/" << gcmds[i].cmd << endl << "    " << gcmds[i].desc << endl;
}


/** broadcast_cmd
 *
 * Sends a message to all users logged on (if they are running p4im). Called
 * when a user enters any line of text that does not start with the '/'
 * character. Equivalent to entering a message in IRC.
 */
void broadcast_cmd(const string &msg, const string &unused)
{
  // build message
  string tagged = guser + ": " + msg;

  // broadcast message
  cout << tagged;
  p4_broadcast(kp4im_pluginid, tagged.c_str(), tagged.length() + 1, false);
}


/** set_cmd
 *
 * Sets the given setting to the given value.
 */
void set_cmd(const string &key, const string &value)
{
  if (key.length() == 0) {  // no arg, so print all settings and values
    assert(value.length() == 0);
    cout << csettings::instance();
  }

  string sanitized_val(value);

  if (sanitized_val.length() == 0)
    sanitized_val = "";

  csettings::instance().set(key, value);
}


/** open_cmd
 *
 * Opens a new connection to the P4 client at the given addr/port.
 */
void open_cmd(const string &user, const string &addr_port)
{
  if (user.length() == 0 || user.length() > kp4_username_len) {
    cout << "Bad username." << endl;
    return;
  } else if  (!addr_port.find(':')) {
    cout << "Bad address." << endl;
    return;
  }


  // extract address and port
  stringstream apstream;
  char addrbuf[kmsg_max_len];

  apstream.unsetf(ios::skipws);
  apstream << addr_port;

  char space;
  apstream >> space;
  apstream.getline(addrbuf, kmsg_max_len, ':');

  ushort port;
  apstream >> port;


  // connect!
  uint ret = p4_add_connection(user.c_str(), addrbuf, port);
  if (ret != 0)
    cout << "Could not connect, error " << ret << "." << endl;
}


/** load_cmd
 *
 * Adds a new plugin id to the list of plugin ids run by P4IM, and calls
 * plugin_loaded on it.
 */
void load_cmd(const string &plugin_id, const string &unused2)
{
  // extract plugin id
  ushort id = 0;
  stringstream stream;
  stream << plugin_id;
  stream >> id;
  if (id == 0) {
    cout << "Bad plugin id." << endl;
    return;
  }

  // load plugin
  if (find(plugins.begin(), plugins.end(), id) != plugins.end()) {
    cout << "Plugin " << id << " is already loaded." << endl;
    return;
  }

  uint ret = p4_plugin_loaded(id);
  if (ret != kerr_success)
    cout << "Couldn't load plugin, error " << ret << "." << endl;
  else
    plugins.push_back(id);
}


/** unload_cmd
 *
 * Removes a plugin id from the list of plugin ids run by P4IM, and calls
 * plugin_stopped on it.
 */
void unload_cmd(const string &plugin_id, const string &unused2)
{
  // extract plugin id
  ushort id = 0;
  stringstream stream;
  stream << plugin_id;
  stream >> id;
  if (id == 0) {
    cout << "Bad plugin id." << endl;
    return;
  }


  // unload plugin
  vector<ushort>::iterator where = find(plugins.begin(), plugins.end(), id);
  if (where == plugins.end()) {
    cout << "Plugin " << id << " is not loaded." << endl;
    return;
  }

  uint ret = p4_plugin_stopped(id);
  if (ret != kerr_success)
    cout << "Couldn't unload plugin, error " << ret << "." << endl;
  else
    plugins.erase(where);
}


/** plugins_cmd
 *
 * Lists the plugin ids that are currently loaded.
 */
void plugins_cmd(const string &unused1, const string &unused2)
{
  vector<ushort>::const_iterator cit;
  for (cit = plugins.begin(); cit != plugins.end(); cit++)
    cout << *cit << endl;
}


/*****************************************************************************
 * HELPERS
 *****************************************************************************/
void error(const string &err)
{
  cerr << err << endl;
  exit(-1);
}


/** parse
 *
 * Takes the given line of user input and passes it to the appropriate command
 * function.
 */
void parse(const string &line)
{
  if (line[0] != '/') {
    broadcast_cmd(line);
    return;
  }

  string cmdstr, arg1;
  char slash, buf[kmsg_max_len];

  istringstream in(line.c_str());
  in >> slash >> cmdstr >> arg1;

  in.getline(buf, 256);        // since arg2 could have spaces in it
  string arg2(buf);

  uint i;
  for (i = 0; i < gcmds.size(); i++) {
    if (cmdstr == gcmds[i].cmd) {
      gcmds[i].fn(arg1, arg2);
      return;
    }
  }

  cout << "Unrecognized command " << cmdstr
       << ". Type /help for a list of commands.";
}


